unit MyExt;

interface
uses QSObjs, QSheme, QSTypes, MyAcc;

type
  TMan = class;

  TPC = class(TChannel)
  private
    FGetParcelTime : TCustTime;
    procedure PCAfterTake(Sender : TElement);
    procedure PCBeforeDrop(Sender: TElement);
    procedure ClearContainer; override;
    function GetLoadFactor : single;
  public
    WorkingTime : TCustTime;
    Queue : TMyAccum;
    constructor Create; override;
    property LoadFactor: single read GetLoadFactor;
  end;


  TMind = class(TChannel)
  private
    procedure MindAfterTake(Sender : TElement);
    procedure MindAfterDrop(Sender : TElement);
    function IsValidSource(Source: TElement): Boolean;
  public
    Man : TMan;
    constructor Create; override;
  end;

  TArms = class(TSource)
  public
    StringCount: Integer;
    Man : TMan;
    constructor Create; override;
    procedure ArmAfterDrop(Sender : TElement);
    procedure ArmBeforeDrop(Sender : TElement);
    function GetCanTake: Boolean; override;
    procedure Pass(SourceIndex : integer); override;
    procedure ClearContainer; override;
    procedure Work; override;
  end;

  TMan = class
  private
    FWaiting : Boolean;
    function GetProbablyWait: single;
  public
    ParcelsPassed : integer;
    ParcelsWaiting : integer;
    Arms : TArms;
    Mind : TMind;
    property Waiting : Boolean read FWaiting write FWaiting;
    property ProbablyWait : single read GetProbablyWait;
  end;

  TMyParcel = class(TParcel)
  public
    Waiting : Boolean;
  end;

implementation


constructor TPC.Create;
begin
  inherited;
  AfterTake := PCAfterTake;
  BeforeDrop := PCBeforeDrop;
end;

procedure TPC.ClearContainer;
begin
  inherited;
  WorkingTime := 0;
end;

procedure TPC.PCAfterTake(Sender : TElement);
var i : integer;
begin
  if Assigned(Queue) and not (Queue.Empty) then begin
    for i := 0 to Queue.ParcelCount-1 do
      TMyParcel(Queue.Parcels[i]).Waiting := True;
  end;
  FGetParcelTime := TQSheme(Sheme).SysTime;
end;

procedure TPC.PCBeforeDrop(Sender : TElement);
begin
{$ifdef Precision}           
  Inc(WorkingTime,EndWorkTime-FGetParcelTime);
{$else}
  Inc(WorkingTime,TQSheme(Sheme).SysTime-FGetParcelTime);
{$endif}
end;

function TPC.GetLoadFactor: Single;
begin
  Result := WorkingTime / TQSheme(Sheme).SysPeriod;
end;

function TMind.IsValidSource(Source : TElement): Boolean;
begin
  Result := Source.Container.Info = Id;
end;

constructor TMind.Create;
begin
  inherited;
  AfterDrop := MindAfterDrop;
  AfterTake := MindAfterTake;
  OnSourceValidate := IsValidSource;
end;

procedure TMind.MindAfterTake(Sender : TElement);
begin
  if TMyParcel(Container).Waiting then Inc(Man.ParcelsWaiting);
  Inc(Man.ParcelsPassed);
end;

procedure TMind.MindAfterDrop(Sender : TElement);
begin
  Man.Waiting := False;
end;

constructor TArms.Create;
begin
  inherited;
  AfterDrop := ArmAfterDrop;
  BeforeDrop := ArmBeforeDrop;
end;

procedure TArms.ClearContainer;
begin
  inherited;
  Man.Waiting := False;
  Man.ParcelsWaiting := 0;
  Man.ParcelsPassed := 0;
end;

function TArms.GetCanTake: Boolean;
begin
  Result := (not Man.Waiting) and inherited GetCanTake;
end;

procedure TArms.ArmBeforeDrop(Sender : TElement);
begin
  TMyParcel(Container).Waiting := False;
end;

procedure TArms.ArmAfterDrop(Sender: TElement);
begin
  Man.Waiting := True;
end;

procedure TArms.Pass(SourceIndex : integer);
begin
  inherited;
  Container.Info := Id;
end;

procedure TArms.Work;
var i : integer;
    Delta : TCustTime;
begin
  Delta := 0;
  for i := 0 to StringCount - 1 do Inc(Delta,Generator.Time);
  EndWorkTime := TQSheme(Sheme).SysTime + Delta;
end;

function TMan.GetProbablyWait: single;
begin
  if ParcelsPassed <> 0 then Result := ParcelsWaiting / ParcelsPassed
  else Result := 1;
end;

end.
